#!/bin/bash
# 
# Execute this script to generate a configure script
#
# This file is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.
#
plugin=fino
majorversion=0.3

# check if we are using a version control system
if [ -d ".bzr" ]; then 
 if [ "`which bzr`" = "" ]; then
  echo "this tree uses bazaar but bzr is not installed" 
  exit 1
 fi
 vcs=bzr
elif [ -d ".hg" ]; then
 if [ "`which hg`" = "" ]; then
  echo "this tree uses mercurial but hg is not installed" 
  exit 1
 fi
 vcs=hg
fi

# locate wasora source dir, first
if [ ! -z "${1}" ]; then
 WASORA_PATH=${1}
elif [ ! -z "${WASORA_DIR}" ]; then
 WASORA_PATH=${WASORA_DIR}
else
 # the "trunk" part is a legacy from bazaar (which itself is a legacy from svn)
 TESTDIRS="../wasora ../../wasora ../../../wasora ../wasora/trunk ../../wasora/trunk ../../../wasora/trunk"
 for i in ${TESTDIRS}; do
  if [[ -z "${WASORA_PATH}" ]] && [[ -d $i ]]; then 
   WASORA_PATH=$i
  fi
 done
 if [ -z "${WASORA_PATH}" ]; then
  echo << EOF
cannot find wasora trunk tree in
${TESTDIRS}
please either set the WASORA_DIR environment variable
or give $0 an argument where the wasora source tree is.
See the INSTALL file for further details.
EOF
  exit 1
 fi
fi

echo "using wasora source from "${WASORA_PATH}
if [ ! -e "${WASORA_PATH}/src/wasora.h" ]; then
  echo "error: cannot find ${WASORA_PATH}/src/wasora.h"
  exit 1
fi

echo -n "cleaning... "
sh autoclean.sh > /dev/null 2>&1
echo "done"

# plugin version
if [ "${vcs}" = "bzr" ]; then
 echo -n "getting bzr revision id... "
 version="${majorversion}.`bzr version-info --custom --template="{revno}"`"
 bzr version-info --custom --template="\
#define PLUGIN_VCS_MAJOR     \"${majorversion}\"\n\
#define PLUGIN_VCS_BRANCH    \"{branch_nick}\"\n\
#define PLUGIN_VCS_REVID     \"{revision_id}\"\n\
#define PLUGIN_VCS_REVNO     {revno}\n\
#define PLUGIN_VCS_DATE      \"{date}\"\n\
#define PLUGIN_VCS_CLEAN     {clean}\n" > src/versionvcs.h
 echo "done"
 echo -n "building changelog... "
 bzr log --gnu-changelog > ChangeLog
 echo "done"

elif [ "${vcs}" = "hg" ]; then
 echo -n "getting hg revision id... "
 version="${majorversion}.`hg log -r tip --template="{rev}"`"
 hg log -r tip --template="\
#define PLUGIN_VCS_MAJOR     \"${majorversion}\"\n\
#define PLUGIN_VCS_BRANCH    \"{branch}\"\n\
#define PLUGIN_VCS_REVID     \"{node}\"\n\
#define PLUGIN_VCS_REVNO     {rev}\n\
#define PLUGIN_VCS_DATE      \"{date|isodate}\"\n" > src/versionvcs.h
# clean tag?
 echo "done"
 echo -n "building changelog... "
 hg log --style changelog > ChangeLog
 echo "done"

else
 version=${majorversion}
 touch ChangeLog
fi


echo -n "formatting readme & install... "
if [ ! -z "`which pandoc`" ]; then
 pandoc README.md -o README
 pandoc README.md -o README.pdf
 pandoc README.md -o README.html -s -S
 pandoc INSTALL.md -o INSTALL
 pandoc INSTALL.md -o INSTALL.pdf
 pandoc INSTALL.md -o INSTALL.html -s -S
 pandoc ${WASORA_PATH}/PLUGINS.md -o PLUGINS
 pandoc ${WASORA_PATH}/PLUGINS.md -o PLUGINS.pdf
 pandoc ${WASORA_PATH}/PLUGINS.md -o PLUGINS.html -s -S
else
 fmt -s README.md > README
 fmt -s INSTALL.md > INSTALL
 fmt -s ${WASORA_PATH}/PLUGINS.md > PLUGINS
fi
echo "done"

# configure.ac
echo -n "building configure.ac..."
cat << EOF > configure.ac
dnl Process this file with autoconf to produce a configure script.
dnl
dnl This file is free software; as a special exception the author gives
dnl unlimited permission to copy and/or distribute it, with or without
dnl modifications, as long as this notice is preserved.
dnl
dnl This program is distributed in the hope that it will be useful, but
dnl WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
dnl implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

AC_INIT([${plugin}],[${version}],[gtheler@cites-gss.com])
AC_CONFIG_SRCDIR([src/${plugin}.c])

AC_CONFIG_MACRO_DIR([m4])

# check for disable-standalone (default is enable)
AC_ARG_ENABLE([standalone],
   [AS_HELP_STRING([--disable-standalone], [do not compile standalone executable @<:@default=enable@:>@])],
   [build_standalone=\${enableval}],
   [build_standalone=yes])

# check for disable-plugin (default is enable)
AC_ARG_ENABLE([plugin],
   [AS_HELP_STRING([--disable-plugin], [do not compile plugin @<:@default=enable@:>@])],
   [build_plugin=\${enableval}],
   [build_plugin=yes])

# if we need to build standalone executable, set the
#  am conditinal and look for libraries
AM_CONDITIONAL([WANT_STANDALONE], [test "x\$build_standalone" != xno])

AS_IF([test "x\$build_standalone" != xno],
        [AC_ENABLE_STATIC],
        [AC_DISABLE_STATIC])
AS_IF([test "x\$build_plugin" != xno],
        [AC_ENABLE_SHARED],
        [AC_DISABLE_SHARED])

WASORA_PLUGIN_INIT_C
WASORA_OPT_FLAGS_C

WASORA_CHECK_WASORA_DIR

AC_CONFIG_FILES([Makefile src/Makefile])

WASORA_PLUGIN_VERSION_H

AS_IF([test "x\$build_standalone" != xno],
        [WASORA_CHECK_BASIC_HEADERS
         WASORA_CHECK_BASIC_LIBS
         WASORA_CHECK_GSL
         WASORA_CHECK_IDA
         WASORA_CHECK_READLINE
         AC_CONFIG_FILES([src/wasora/Makefile])])

# fino needs petsc and optionally slepc
WASORA_CHECK_PETSC([required])
WASORA_CHECK_SLEPC([optional])

AC_OUTPUT

AS_IF([test "x\$build_standalone" != xno], [
  echo
  AS_BOX([Configuration summary for ${plugin} standalone executable])
  WASORA_RESUME_LIBS
])

WASORA_RESUME_PETSC
WASORA_RESUME_SLEPC

# para saber que incluir en el makefile, si PETSc solito o SLEPc
AC_SUBST([PETSC_DIR])
AC_SUBST([PETSC_ARCH])
AC_SUBST([SLEPC_DIR])
AS_IF([test \$slepc_found -ne 0],[
  echo "include \\\$(SLEPC_DIR)/conf/slepc_variables" > petscslepc.mak
],[
  echo "include \\\$(PETSC_DIR)/conf/variables" > petscslepc.mak
])
# listorti


AS_IF([ test \$ida = 0 -o \$readline = 0 ], [
  WASORA_FINAL_WARN([0],[1])
],[
  WASORA_FINAL_WARN([1],[1])
])

EOF
echo "done"

echo -n "copying wasora source... "
rm -rf src/wasora
mkdir src/wasora
cp -r ${WASORA_PATH}/src/* src/wasora
rm -f src/wasora/wasora src/wasora/Makefile* src/wasora/*.o  src/wasora/*.lo src/wasora/*.lo
if [ ! -f src/wasora/version.h ]; then
 echo "error: the wasora tree at ${WASORA_PATH} is not autoconfigured yet"
 exit 1
fi
echo "done"

am="src/wasora/Makefile.am"
echo -n "building $am... "
cat << EOF > $am
include ../../petscslepc.mak

AUTOMAKE_OPTIONS = subdir-objects
ACLOCAL_AMFLAGS = \${ACLOCAL_FLAGS}

bin_PROGRAMS = ${plugin}

${plugin}_SOURCES = \\
EOF
cd src/wasora
find . -maxdepth 2 \( -name "*.c" -o -name "*.h" \) | xargs >> ../../$am
cd ../..

cat << EOF >> $am

${plugin}_INCLUDES = \$(WASORA_INCLUDE) \$(SLEPC_CC_INCLUDES) \$(PETSC_CC_INCLUDES) \$(CC_INCLUDES) \$(all_includes)
${plugin}_LDADD = ../.libs/lib${plugin}.a \$(STANDALONELIBS) \$(SLEPC_LIB) \$(PETSC_LIB) \$(all_libraries)

${plugin}_CFLAGS = -DHARDCODEDPLUGIN
${plugin}_LDFLAGS =

EOF
echo "done"

# plugin
am="src/Makefile.am"
echo -n "building $am... "
cat << EOF > $am
include ../petscslepc.mak

AUTOMAKE_OPTIONS = subdir-objects
ACLOCAL_AMFLAGS = \$(ACLOCAL_FLAGS)

# no se por que no camina como el de arriba
DEFAULT_INCLUDES = \$(WASORA_INCLUDE) \$(SLEPC_CC_INCLUDES) \$(PETSC_CC_INCLUDES) \$(CC_INCLUDES)

lib_LTLIBRARIES = lib${plugin}.la
# lib${plugin}_la_LDFLAGS = -no-undefined
lib${plugin}_la_INCLUDES = \$(all_includes)
lib${plugin}_la_LIBADD = \$(SLEPC_LIB) \$(PETSC_LIB)
lib${plugin}_la_SOURCES = ./version.h \\
EOF

cd src
find . mesh -maxdepth 1 \( -name "*.c" -o -name "*.h" \) | xargs >> ../$am
cd ..
echo "done"


echo "calling autoreconf... "
mkdir -p ./m4
cp ${WASORA_PATH}/m4/*.m4 ./m4
touch petscslepc.mak
autoreconf -i
echo "done"

