/*------------ -------------- -------- --- ----- ---   --       -            -
 *  fino's construction of linear elastic problem (break) with optional vibration (shake)
 *
 *  Copyright (C) 2015 jeremy theler
 *
 *  This file is part of fino.
 *
 *  fino is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fino is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>

#include "fino.h"

#undef  __FUNCT__
#define __FUNCT__ "fino_build_breakshake"
int fino_build_breakshake(element_t *element, int v) {

  static int J;            // cantidad de nodos locales
  static fino_distribution_t E;
  static fino_distribution_t nu;
  static fino_distribution_t rho;
  static fino_distribution_t fx;
  static fino_distribution_t fy;
  static fino_distribution_t fz;
  
  
  // matrices de la formulacion del problema
  static gsl_matrix *C = NULL;
  static gsl_matrix *B = NULL;

  // matriz intermedias
  static gsl_matrix *CB;

  double c;
  
  double w_gauss;
  int j;

  PetscFunctionBegin;

  w_gauss = mesh_compute_fem_objects_at_gauss(fino.mesh, element, v); 
  
  // TODO: improve
  // matriz C de la formulacion
  if (C == NULL) {
    wasora_call(fino_distribution_init(&E, "E"));
    wasora_call(fino_distribution_init(&nu, "nu"));
    if (fino.problem == problem_shake) {
      wasora_call(fino_distribution_init(&rho, "rho"));
    } else if (fino.problem == problem_break) {
      wasora_call(fino_distribution_init(&fx, "fx"));
      wasora_call(fino_distribution_init(&fy, "fy"));
      wasora_call(fino_distribution_init(&fz, "fz"));
    }
    
    C = gsl_matrix_calloc(6, 6);
    
    // si E y nu son variables, calculamos C una sola vez y ya
    if (E.variable != NULL && nu.variable != NULL) {
      wasora_call(fino_break_compute_C(C, &E, &nu));
    }
  }
  
  if (J != element->type->nodes) {
    J = element->type->nodes;
    gsl_matrix_free(B);
    B = gsl_matrix_alloc(6, 3*J);
    
    gsl_matrix_free(CB);
    CB = gsl_matrix_alloc(6, 3*J);
  }  
  
  // la H es la misma, pero la B no es la misma que la fem porque la formulacion es reducida
  gsl_matrix_set_zero(B);

  for (j = 0; j < J; j++) {
    gsl_matrix_set(B, 0, 3*j+0, gsl_matrix_get(fino.mesh->fem.dhdx, j, 0));
    gsl_matrix_set(B, 1, 3*j+1, gsl_matrix_get(fino.mesh->fem.dhdx, j, 1));
    gsl_matrix_set(B, 2, 3*j+2, gsl_matrix_get(fino.mesh->fem.dhdx, j, 2));
    
    gsl_matrix_set(B, 3, 3*j+0, gsl_matrix_get(fino.mesh->fem.dhdx, j, 1));
    gsl_matrix_set(B, 3, 3*j+1, gsl_matrix_get(fino.mesh->fem.dhdx, j, 0));

    gsl_matrix_set(B, 4, 3*j+1, gsl_matrix_get(fino.mesh->fem.dhdx, j, 2));
    gsl_matrix_set(B, 4, 3*j+2, gsl_matrix_get(fino.mesh->fem.dhdx, j, 1));

    gsl_matrix_set(B, 5, 3*j+0, gsl_matrix_get(fino.mesh->fem.dhdx, j, 2));
    gsl_matrix_set(B, 5, 3*j+2, gsl_matrix_get(fino.mesh->fem.dhdx, j, 0));
    
    if (fino.problem == problem_break) {
      // el vector de fuerzas volumetricas
      c = w_gauss * gsl_vector_get(fino.mesh->fem.h, j);
      gsl_vector_add_to_element(fino.bi, 3*j+0, c * fino_distribution_evaluate(&fx));
      gsl_vector_add_to_element(fino.bi, 3*j+1, c * fino_distribution_evaluate(&fy));
      gsl_vector_add_to_element(fino.bi, 3*j+2, c * fino_distribution_evaluate(&fz));
    }
    
  }

  // si E y nu estan dadas por variables, C es constante y no la volvemos a evaluar
  // pero si alguna es una funcion, es otro cantar
  // TODO: ver si se justificar hacer algo parecido para las otras distribuciones
  if (E.function != NULL || nu.function != NULL) {
    wasora_call(fino_break_compute_C(C, &E, &nu));
  }

  // calculamos Bt*C*B
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, C, B, 0, CB);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, w_gauss, B, CB, 1.0, fino.Ai);
  
  if (fino.problem == problem_shake) {
    // calculamos Ht*rho*H
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, w_gauss * fino_distribution_evaluate(&rho), fino.mesh->fem.H, fino.mesh->fem.H, 1.0, fino.Bi);
  } 
  
  PetscFunctionReturn(WASORA_RUNTIME_OK);
  
}

#undef  __FUNCT__
#define __FUNCT__ "fino_break_compute_C"
int fino_break_compute_C(gsl_matrix *C, fino_distribution_t *E, fino_distribution_t *nu) {
  
  double evaluated_nu;
  double c1, c2, c3, c4;
  
  PetscFunctionBegin;

  evaluated_nu = fino_distribution_evaluate(nu);
  c1 = fino_distribution_evaluate(E)/((1+evaluated_nu)*(1-2*evaluated_nu));
  c2 = c1 * evaluated_nu;
  c3 = c1 * (1-evaluated_nu);
  c4 = c1 * (1-2*evaluated_nu)/2;
    
  gsl_matrix_set(C, 0, 0, c3);
  gsl_matrix_set(C, 0, 1, c2);
  gsl_matrix_set(C, 0, 2, c2);

  gsl_matrix_set(C, 1, 0, c2);
  gsl_matrix_set(C, 1, 1, c3);
  gsl_matrix_set(C, 1, 2, c2);

  gsl_matrix_set(C, 2, 0, c2);
  gsl_matrix_set(C, 2, 1, c2);
  gsl_matrix_set(C, 2, 2, c3);
  
  gsl_matrix_set(C, 3, 3, c4);
  gsl_matrix_set(C, 4, 4, c4);
  gsl_matrix_set(C, 5, 5, c4);

  PetscFunctionReturn(WASORA_RUNTIME_OK);
}    



#undef  __FUNCT__
#define __FUNCT__ "fino_break_compute_vonmisses"
int fino_break_compute_vonmisses(void) {
  
  double E, nu;
  double ex, ey, ez;
  double gammaxy, gammayz, gammazx;
  double sigmax, sigmay, sigmaz;
  double tauxy, tauyz, tauzx;
  double c1, c2;
  int i;
  
  PetscFunctionBegin;
  
  fino.sigma->data_argument = fino.gradsol[0][0]->data_argument;
  free(fino.sigma->data_value);
  fino.sigma->data_size = fino.mesh->n_cells;
  fino.sigma->data_value = calloc(fino.mesh->n_cells, sizeof(double));

  E = wasora_get_var_value("E");
  nu = wasora_get_var_value("nu");
  c1 = E/((1+nu)*(1-2*nu));
  c2 = (1-2*nu)/2;
  wasora_var(fino.vars.sigma_max) = 0;
  
  for (i = 0; i < fino.mesh->n_cells; i++) {

/*  
e_x(x,y,z) := du_xdx(x,y,z)
e_y(x,y,z) := du_ydy(x,y,z)
e_z(x,y,z) := du_zdz(x,y,z)
gamma_xy(x,y,z) := du_xdy(x,y,z) + du_ydx(x,y,z)
gamma_yz(x,y,z) := du_ydz(x,y,z) + du_zdy(x,y,z)
gamma_zx(x,y,z) := du_zdx(x,y,z) + du_xdz(x,y,z)

sigma_x(x,y,z) := E/((1+nu)*(1-2*nu))*((1-nu)*e_x(x,y,z) + nu*(e_y(x,y,z)+e_z(x,y,z)))
sigma_y(x,y,z) := E/((1+nu)*(1-2*nu))*((1-nu)*e_y(x,y,z) + nu*(e_x(x,y,z)+e_z(x,y,z)))
sigma_z(x,y,z) := E/((1+nu)*(1-2*nu))*((1-nu)*e_z(x,y,z) + nu*(e_x(x,y,z)+e_y(x,y,z)))
tau_xy(x,y,z) := E/((1+nu)*(1-2*nu))*(1-2*nu)/2*gamma_xy(x,y,z)
tau_yz(x,y,z) := E/((1+nu)*(1-2*nu))*(1-2*nu)/2*gamma_yz(x,y,z)
tau_zx(x,y,z) := E/((1+nu)*(1-2*nu))*(1-2*nu)/2*gamma_zx(x,y,z)
VM_stress(x,y,z) := sqrt(1/2*((sigma_x(x,y,z)-sigma_y(x,y,z))^2 + (sigma_y(x,y,z)-sigma_z(x,y,z))^2 + (sigma_z(x,y,z)-sigma_x(x,y,z))^2 + 6*(tau_xy(x,y,z)^2+tau_yz(x,y,z)^2+tau_zx(x,y,z)^2)))
*/
    
    ex = fino.gradsol[0][0]->data_value[i];
    ey = fino.gradsol[1][1]->data_value[i];
    ez = fino.gradsol[2][2]->data_value[i];
    gammaxy = fino.gradsol[0][1]->data_value[i] + fino.gradsol[1][0]->data_value[i];
    gammayz = fino.gradsol[1][2]->data_value[i] + fino.gradsol[2][1]->data_value[i];
    gammazx = fino.gradsol[2][0]->data_value[i] + fino.gradsol[0][2]->data_value[i];
    
    sigmax = c1 * ((1-nu)*ex + nu*(ey+ez));
    sigmay = c1 * ((1-nu)*ey + nu*(ex+ez));
    sigmaz = c1 * ((1-nu)*ez + nu*(ex+ey));
    tauxy =  c1 * c2 * gammaxy;
    tauyz =  c1 * c2 * gammayz;
    tauzx =  c1 * c2 * gammazx;
    
    if ((fino.sigma->data_value[i] = sqrt(0.5*(gsl_pow_2(sigmax-sigmay) +
                                              gsl_pow_2(sigmay-sigmaz) +
                                              gsl_pow_2(sigmaz-sigmax) +
                                              6 * (gsl_pow_2(tauxy) + gsl_pow_2(tauyz) + gsl_pow_2(tauzx))))
        ) > wasora_var(fino.vars.sigma_max)) {
      wasora_var(fino.vars.sigma_max) = fino.sigma->data_value[i];
      wasora_var(fino.vars.sigma_max_x) = fino.mesh->cell[i].x[0];
      wasora_var(fino.vars.sigma_max_y) = fino.mesh->cell[i].x[1];
      wasora_var(fino.vars.sigma_max_z) = fino.mesh->cell[i].x[2];
    }
    
  }
  
  PetscFunctionReturn(WASORA_RUNTIME_OK);
}
