/*------------ -------------- -------- --- ----- ---   --       -            -
 *  fino's initialization routines
 *
 *  Copyright (C) 2015 jeremy theler
 *
 *  This file is part of fino.
 *
 *  fino is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fino is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#include <signal.h>
#include "fino.h"

#define NAME_SIZE 32

int plugin_init_before_parser(void) {

  char *dummy;
  int i;
  
  if (sizeof(PetscReal) != sizeof(double)) {
    wasora_push_error_message("PETSc should be compiled with double-precision real scalar types");
    return WASORA_PARSER_ERROR;
  }
  
  // amasamos la linea de comandos original (porque la que saca getopt puede tener un orden que no nos sirve)
  // el chiste es que por ejemplo "-log_summary" es atrapado por el getopt de wasora como "-l"
  // hay que re-escrbir eso como "--slepc_opt log_summary"
  // si alguna opcion tiene argumento hay que ponerlo como "--slepc_opt pc_type=sor"
  for (i = 0; i < wasora.argc_orig; i++) {
    if (strcmp(wasora.argv_orig[i], "--petsc_opt") == 0) {
      if (i >= (wasora.argc_orig-1)) {
        wasora_push_error_message("commandline option --petsc_opt needs an argument");
        return WASORA_PARSER_ERROR;
      } else if (wasora.argv_orig[i+1][0] == '-') {
        wasora_push_error_message("the argument of commandline option --petsc_opt should not start with a dash (it is added automatically)");
        return WASORA_PARSER_ERROR;
      }
      
      if ((dummy = strchr(wasora.argv_orig[i+1], '=')) != NULL)  {
        char *tmp1, *tmp2;
        *dummy = '\0';
        tmp1 = strdup(wasora.argv_orig[i+1]);
        tmp2 = strdup(dummy+1);
        wasora.argv_orig[i]   = realloc(wasora.argv_orig[i],   strlen(wasora.argv_orig[i+1])+2);
        wasora.argv_orig[i+1] = realloc(wasora.argv_orig[i+1], strlen(dummy)+1);
        sprintf(wasora.argv_orig[i],  "-%s", tmp1);
        sprintf(wasora.argv_orig[i+1], "%s", tmp2);
        free(tmp1);
        free(tmp2);
        
      } else {
        char *tmp1;
        tmp1 = strdup(wasora.argv_orig[i+1]);
        wasora.argv_orig[i+1] = realloc(wasora.argv_orig[i+1], strlen(tmp1)+1);
        wasora.argv_orig[i][0] = '\0';
        sprintf(wasora.argv_orig[i+1],  "-%s", tmp1);
        free(tmp1);
      }
      i++;
    }
  }

#ifdef HAVE_SLEPC  
  // inicializamos la slepc (que a su vez inicializa la petsc)
  // le pasamos la linea de comandos que acabamos de amasar
  petsc_call(SlepcInitialize(&wasora.argc_orig, &wasora.argv_orig, (char*)0, PETSC_NULL));
#else
  // inicializamos la petsc
  // le pasamos la linea de comandos que acabamos de amasar
  petsc_call(PetscInitialize(&wasora.argc_orig, &wasora.argv_orig, (char*)0, PETSC_NULL));
#endif
  fino.petscinit_called = 1;
  
  // los segfaults son segfaults, no queremos que la petsc meta las narices
  signal(SIGSEGV, SIG_DFL);

  // esto lo vamos a usar despues cuando hagamos el chiste en paralelo
  MPI_Comm_rank(PETSC_COMM_WORLD, &fino.rank);
  MPI_Comm_size(PETSC_COMM_WORLD, &fino.size);

  // instalamos nuestro error handler para errores la petsc 
  PetscPushErrorHandler(&fino_handler, NULL);

  // inicializamos mesh
  wasora_call(wasora_mesh_init_before_parser());
  // como somos un codigo de elementos finitos no aseguramos de que
  // el defaults sea trabajar sobre nodos
  wasora_mesh.default_cell_centered = 0;

  
  // variables especiales de fino
  // TODO: documentar
  fino.vars.atol = wasora_define_variable("fino_atol");
  wasora_var(fino.vars.atol) = 1e-50;
  fino.vars.rtol = wasora_define_variable("fino_rtol");
  wasora_var(fino.vars.rtol) = 1e-9;
  fino.vars.divtol = wasora_define_variable("fino_divtol");
  wasora_var(fino.vars.divtol) = 1e+3;
  fino.vars.max_it = wasora_define_variable("fino_max_it");
  wasora_var(fino.vars.max_it) = 1e4;

  fino.vars.dirichlet_diagonal = wasora_define_variable("fino_dirichlet_diagonal");
  
  fino.vars.nodes = wasora_define_variable("nodes");
  fino.vars.elements = wasora_define_variable("elements");
  
  // estas dos se las pedimos a petsc por si al usuario le interesa
  fino.vars.n_it = wasora_define_variable("fino_n_it");
  fino.vars.residual= wasora_define_variable("fino_residual");

  fino.vars.sigma_max = wasora_define_variable("sigma_max");
  fino.vars.sigma_max_x = wasora_define_variable("sigma_max_x");
  fino.vars.sigma_max_y = wasora_define_variable("sigma_max_y");
  fino.vars.sigma_max_z = wasora_define_variable("sigma_max_z");
  
  // variables internas
///va+lambda+name lambda
///va+lambda+desc 
///va+lambda+desc Requested eigenvalue. It is equal to 1.0 until
///va+lambda+desc `FINO_STEP` is executed.  
  fino.vars.lambda = wasora_define_variable("lambda");
  wasora_var(fino.vars.lambda) = 1.0;
  
///va+time_wall_build+name time_wall_build
///va+time_wall_build+desc Wall time insumed to build the problem matrices, in seconds.
  fino.vars.time_wall_build = wasora_define_variable("time_wall_build");

///va+time_wall_solve+name time_wall_solve
///va+time_wall_solve+desc Wall time insumed to solve the eigen-problem, in seconds.
  fino.vars.time_wall_solve = wasora_define_variable("time_wall_solve");

///va+time_wall_total+name time_wall_total
///va+time_wall_total+desc Wall time insumed to initialize, build and solve, in seconds.
  fino.vars.time_wall_total = wasora_define_variable("time_wall_total");
  
///va+time_cpu_build+name time_cpu_build
///va+time_cpu_build+desc CPU time insumed to build the problem matrices, in seconds.
  fino.vars.time_cpu_build = wasora_define_variable("time_cpu_build");

///va+time_cpu_solve+name time_cpu_solve
///va+time_cpu_solve+desc CPU time insumed to solve the eigen-problem, in seconds.
  fino.vars.time_cpu_solve = wasora_define_variable("time_cpu_solve");

///va+time_wall_total+name time_cpu_total
///va+time_wall_total+desc CPU time insumed to initialize, build and solve, in seconds.
  fino.vars.time_cpu_total = wasora_define_variable("time_cpu_total");
  
///va+time_petsc_build+name time_petsc_build
///va+time_petsc_build+desc CPU time insumed by PETSc to build the problem matrices, in seconds.
  fino.vars.time_petsc_build = wasora_define_variable("time_petsc_build");

///va+time_petsc_solve+name time_petsc_solve
///va+time_petsc_solve+desc CPU time insumed by PETSc to solve the eigen-problem, in seconds.
  fino.vars.time_petsc_solve = wasora_define_variable("time_petsc_solve");

///va+time_wall_total+name time_wall_total
///va+time_wall_total+desc CPU time insumed by PETSc to initialize, build and solve, in seconds.
  fino.vars.time_petsc_total = wasora_define_variable("time_petsc_total");

  ///va+petsc_flops+name petsc_flops
///va+petsc_flops+desc Number of floating point operations performed by PETSc/SLEPc.
  fino.vars.flops_petsc = wasora_define_variable("flops_petsc");
         
///va+memory_use+name available_memory
///va+memory_use+desc Total available memory, in bytes.
  fino.vars.available_memory = wasora_define_variable("available_memory");

///va+memory_usage_global+name global_memory_use
///va+memory_usage_global+desc Maximum resident set size (global memory used), in bytes.
  fino.vars.memory_usage_global = wasora_define_variable("memory_usage_global");
  
///va+memory_usage_petsc+name petsc_memory_use
///va+memory_usage_petsc+desc Maximum resident set size (memory used by PETSc), in bytes.
  fino.vars.memory_usage_petsc = wasora_define_variable("memory_usage_petsc");
  
  // vector con las funciones de forma
  // eventualmente despues hacemos un free y lo definimos de nuevo
  fino.h = wasora_define_vector("h", J_MAX, NULL);

  // jacobianos
  // eventualmente despues hacemos un free y lo definimos de nuevo
  fino.dhdx = wasora_define_matrix("dhdx", J_MAX, M_MAX, NULL);

  // nombres de objetos por default
  // TODO: ponerlo en otro lado poque si no se necesitan se usan nombres copados
  // que podrian estar disponibles para otra cosa (por ejemplo coef. de heat transfer)
  fino.shape_name = strdup("h");
  fino.lhs_matrix_name = strdup("A");
  fino.rhs_matrix_name = strdup("B");
  fino.rhs_vector_name = strdup("b");
  
  return WASORA_PARSER_OK;
}

int plugin_init_after_parser(void) {
  
  return WASORA_RUNTIME_OK;
}


int plugin_init_before_run(void) {

//  wasora_var(fino.vars.keff) = 1.0;

  fino.problem_size = 0;
  fino.spatial_unknowns = 0;

  wasora_call(fino_problem_free());
  
  return WASORA_RUNTIME_OK;
}


int plugin_finalize(void) {

  if (fino.petscinit_called) {
    if (fino.problem_type == problem_linear) {
      petsc_call(PetscFinalize());
    } else if (fino.problem_type == problem_eigen) {
      petsc_call(SlepcFinalize());
    }
  }
  
  return WASORA_RUNTIME_OK;
}


// esto viene despues de haber leido la malla
int fino_problem_init(void) {

  char name[NAME_SIZE];
  int i, j, k, g, m;
  int width, max_size;

//---------------------------------
// inicializamos parametros
//---------------------------------

  if ((fino.mesh = wasora_mesh.meshes) == NULL) {
    wasora_push_error_message("no mesh");
    return WASORA_RUNTIME_ERROR;
  }
  
  // ponemos esto para hacer explicito que somos FEM y no FVM
  fino.spatial_unknowns = fino.mesh->n_nodes;
  fino.mesh->data_type = data_type_node;
  wasora_var(fino.vars.nodes) = (double)fino.mesh->n_nodes;
  wasora_var(fino.vars.elements) = (double)fino.mesh->n_elements;
  fino.problem_size = fino.spatial_unknowns * fino.degrees;
  
  wasora_call(mesh_init_gauss_points(fino.mesh, fino.mesh->n_gauss_points));

//---------------------------------
// vemos que clase de problema tenemos que resolver
//---------------------------------
  if (fino.problem_type == problem_automatic) {
    char name[NAME_SIZE];
    
    snprintf(name, NAME_SIZE-1, "%s1", fino.rhs_vector_name);
    if (wasora_get_function_ptr(name) != NULL) {
      fino.problem_type = problem_linear;
    }

    snprintf(name, NAME_SIZE-1, "%s1.1", fino.rhs_matrix_name);
    if (wasora_get_function_ptr(name) != NULL) {
      fino.problem_type = problem_eigen;
    }
  }

//---------------------------------
//  linkeamos las funciones que evaluan elementos de objetos elementos a
//  las funciones que dio el usuario (si el problema es generico)
//---------------------------------
  if (fino.problem == problem_generic) {
    max_size = fino.mesh->max_nodes_per_element * fino.degrees;
    fino.Ai_function = calloc(max_size, sizeof(function_t *));
    fino.Bi_function = calloc(max_size, sizeof(function_t *));
    fino.bi_function = calloc(max_size, sizeof(function_t *));
    
    for (i = 0; i < max_size; i++) {
      fino.Ai_function[i] = calloc(max_size, sizeof(function_t));
      if (fino.problem_type == problem_eigen) {
        fino.Bi_function[i] = calloc(max_size, sizeof(function_t));
      }

      for (j = 0; j < max_size; j++) {
        snprintf(name, NAME_SIZE-1, "%s%d.%d", fino.lhs_matrix_name, i+1, j+1);
        if ((fino.Ai_function[i][j] = wasora_get_function_ptr(name)) == NULL) {
           wasora_push_error_message("could not find function '%s'", name);
           return WASORA_RUNTIME_ERROR;
        }

        if (fino.problem_type == problem_eigen) {
          snprintf(name, NAME_SIZE-1, "%s%d.%d", fino.rhs_matrix_name, i+1, j+1);
          if ((fino.Bi_function[i][j] = wasora_get_function_ptr(name)) == NULL) {
             wasora_push_error_message("could not find function '%s'", name);
             return WASORA_RUNTIME_ERROR;
          }
        }
      }

      if (fino.problem_type == problem_linear) {
        snprintf(name, NAME_SIZE-1, "%s%d", fino.rhs_vector_name, i+1);
        if ((fino.bi_function[i] = wasora_get_function_ptr(name)) == NULL) {
           wasora_push_error_message("could not find function '%s'", name);
           return WASORA_RUNTIME_ERROR;
        }
      }
    }
  }
  

//---------------------------------
// alocamos objetos globales
//---------------------------------

  // la matriz global
  width = fino.mesh->max_first_neighbor_nodes * fino.degrees;
  petsc_call(MatCreate(PETSC_COMM_WORLD, &fino.A));
  petsc_call(MatSetSizes(fino.A, PETSC_DECIDE, PETSC_DECIDE, fino.problem_size, fino.problem_size));
  petsc_call(MatSetFromOptions(fino.A));
  petsc_call(MatMPIAIJSetPreallocation(fino.A, width, PETSC_NULL, width, PETSC_NULL));
  petsc_call(MatSeqAIJSetPreallocation(fino.A, width, PETSC_NULL));
  petsc_call(PetscObjectSetName((PetscObject)fino.A, fino.lhs_matrix_name));

  // el vector incognita
  petsc_call(MatGetVecs(fino.A, NULL, &fino.phi));
  petsc_call(PetscObjectSetName((PetscObject)fino.phi, "phi"));

  
  if (fino.problem_type == problem_linear) {
    // el vector del miembro derecho
    petsc_call(MatGetVecs(fino.A, NULL, &fino.b));
    petsc_call(PetscObjectSetName((PetscObject)fino.b, fino.rhs_vector_name));
    
  } else if (fino.problem_type == problem_eigen) {
    // la matriz del miembro derecho para autovalores
    petsc_call(MatCreate(PETSC_COMM_WORLD, &fino.B));
    petsc_call(MatSetSizes(fino.B, PETSC_DECIDE, PETSC_DECIDE, fino.problem_size, fino.problem_size));
    petsc_call(MatSetFromOptions(fino.B));
    petsc_call(MatMPIAIJSetPreallocation(fino.B, width, PETSC_NULL, width, PETSC_NULL));
    petsc_call(MatSeqAIJSetPreallocation(fino.B, width, PETSC_NULL));
    petsc_call(PetscObjectSetName((PetscObject)fino.B, fino.rhs_matrix_name));

  }
  

  fino.mesh->data_argument = calloc(fino.dimensions, sizeof(double *));
  for (m = 0; m < fino.dimensions; m++) {
    fino.mesh->data_argument[m] = calloc(fino.spatial_unknowns, sizeof(double));
  }

  if (fino.mesh->structured) {
    fino.mesh->rectangular_mesh_size[0] = fino.mesh->ncells_x+1;
    fino.mesh->rectangular_mesh_point[0] = calloc(fino.mesh->ncells_x+1, sizeof(double));
    for (i = 0; i < fino.mesh->ncells_x+1; i++) {
      fino.mesh->rectangular_mesh_point[0][i] = fino.mesh->nodes_x[i];
    }
    if (fino.dimensions >= 2) {
      fino.mesh->rectangular_mesh_size[1] = fino.mesh->ncells_y+1;
      fino.mesh->rectangular_mesh_point[1] = calloc(fino.mesh->ncells_y+1, sizeof(double));
      for (j = 0; j < fino.mesh->ncells_y+1; j++) {
        fino.mesh->rectangular_mesh_point[1][j] = fino.mesh->nodes_y[j];
      }
      if (fino.dimensions >= 3) {
        fino.mesh->rectangular_mesh_size[2] = fino.mesh->ncells_z+1;
        fino.mesh->rectangular_mesh_point[2] = calloc(fino.mesh->ncells_z+1, sizeof(double));
        for (k = 0; k < fino.mesh->ncells_z+1; k++) {
          fino.mesh->rectangular_mesh_point[2][k] = fino.mesh->nodes_z[k];
        }
      }
    }
  }
  
  for (k = 0; k < fino.mesh->n_nodes; k++) {
    fino.mesh->data_argument[0][k] = fino.mesh->node[k].x[0];
    if (fino.dimensions >= 2) {
      fino.mesh->data_argument[1][k] = fino.mesh->node[k].x[1];
      if (fino.dimensions == 3) {
        fino.mesh->data_argument[2][k] = fino.mesh->node[k].x[2];
      }
    }
  }
  

  // rellenamos informacion sobre las funciones continuas que van a tener la solucion
  // algo ya pusimos en fino_define_result_functions())
  for (g = 0; g < fino.degrees; g++) {
    fino.solution[g]->data_size = fino.spatial_unknowns;

    if (g == 0) {
      // para el primer grado de libertad fabricamos la malla
      fino.solution[g]->data_argument = malloc(fino.dimensions * sizeof(double *));
      // y ya que estamos apuntamos a las variables argumento
      for (m = 0; m < fino.dimensions; m++) {
        fino.solution[g]->data_argument[m] = malloc(fino.spatial_unknowns * sizeof(double));
      }
      for (k = 0; k < fino.mesh->n_nodes; k++) {
        for (m = 0; m < fino.dimensions; m++) {
          fino.solution[g]->data_argument[m][k] = fino.mesh->node[k].x[m];
        }
      }

    } else {
      // para los otros usamos la que generamos en g = 0
      fino.solution[g]->data_argument = fino.solution[0]->data_argument;
    }

    fino.solution[g]->data_value = calloc(fino.spatial_unknowns, sizeof(double));
    
  }

  return WASORA_PARSER_OK;
}

int fino_problem_free(void) {
  int m, g;

  if (fino.mesh->n_elements != 0) {
    for (m = 0; m < fino.dimensions; m++) {
      free(fino.mesh->data_argument[m]);
    }
    free(fino.mesh->data_argument);

    for (g = 0; g < fino.degrees; g++) {
      free(fino.solution[g]->data_value);
    }
    mesh_free(fino.mesh);
  }
     
  if (fino.phi != PETSC_NULL) {
    petsc_call(VecDestroy(&fino.phi));
  }
  if (fino.A != PETSC_NULL) {
    petsc_call(MatDestroy(&fino.A));
  }
  if (fino.B != PETSC_NULL) {
    petsc_call(MatDestroy(&fino.B));
  }
  if (fino.b != PETSC_NULL) {
    petsc_call(MatDestroy(&fino.B));
  }
  if (fino.ksp != PETSC_NULL) {
    petsc_call(KSPDestroy(&fino.ksp));
  }
  if (fino.eps != PETSC_NULL) {
    petsc_call(EPSDestroy(&fino.eps));
  }
  
  return WASORA_RUNTIME_OK;

}