/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora's mesh-related element routines
 *
 *  Copyright (C) 2015 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#include <wasora.h>
#include "mesh.h"

/*

Line:                   Line3:           Line4:    
                                                
0----------1 --> u      0-----2----1     0----2----3----1

Triangle:               Triangle6:          Triangle9/10:          Triangle12/15:

v                                                              
^                                                                   2 
|                                                                   | \ 
2                       2                    2                      9   8
|`\                     |`\                  | \                    |     \ 
|  `\                   |  `\                7   6                 10 (14)  7
|    `\                 5    `4              |     \                |         \ 
|      `\               |      `\            8  (9)  5             11 (12) (13) 6
|        `\             |        `\          |         \            |             \
0----------1 --> u      0-----3----1         0---3---4---1          0---3---4---5---1

Quadrangle:            Quadrangle8:            Quadrangle9:

      v
      ^
      |
3-----------2          3-----6-----2           3-----6-----2 
|     |     |          |           |           |           | 
|     |     |          |           |           |           | 
|     +---- | --> u    7           5           7     8     5 
|           |          |           |           |           | 
|           |          |           |           |           | 
0-----------1          0-----4-----1           0-----4-----1 

Tetrahedron:                          Tetrahedron10:

                   v
                 .
               ,/
              /
           2                                     2                              
         ,/|`\                                 ,/|`\                          
       ,/  |  `\                             ,/  |  `\       
     ,/    '.   `\                         ,6    '.   `5     
   ,/       |     `\                     ,/       8     `\   
 ,/         |       `\                 ,/         |       `\ 
0-----------'.--------1 --> u         0--------4--'.--------1
 `\.         |      ,/                 `\.         |      ,/ 
    `\.      |    ,/                      `\.      |    ,9   
       `\.   '. ,/                           `7.   '. ,/     
          `\. |/                                `\. |/       
             `3                                    `3        
                `\.
                   ` w

Hexahedron:             Hexahedron20:          Hexahedron27:

       v
3----------2            3----13----2           3----13----2     
|\     ^   |\           |\         |\          |\         |\    
| \    |   | \          | 15       | 14        |15    24  | 14  
|  \   |   |  \         9  \       11 \        9  \ 20    11 \  
|   7------+---6        |   7----19+---6       |   7----19+---6 
|   |  +-- |-- | -> u   |   |      |   |       |22 |  26  | 23| 
0---+---\--1   |        0---+-8----1   |       0---+-8----1   | 
 \  |    \  \  |         \  17      \  18       \ 17    25 \  18
  \ |     \  \ |         10 |        12|        10 |  21    12| 
   \|      w  \|           \|         \|          \|         \| 
    4----------5            4----16----5           4----16----5 

Prism:                      Prism15:               Prism18:

           w
           ^
           |
           3                       3                      3        
         ,/|`\                   ,/|`\                  ,/|`\      
       ,/  |  `\               12  |  13              12  |  13    
     ,/    |    `\           ,/    |    `\          ,/    |    `\  
    4------+------5         4------14-----5        4------14-----5 
    |      |      |         |      8      |        |      8      | 
    |    ,/|`\    |         |      |      |        |    ,/|`\    | 
    |  ,/  |  `\  |         |      |      |        |  15  |  16  | 
    |,/    |    `\|         |      |      |        |,/    |    `\| 
   ,|      |      |\        10     |      11       10-----17-----11
 ,/ |      0      | `\      |      0      |        |      0      | 
u   |    ,/ `\    |    v    |    ,/ `\    |        |    ,/ `\    | 
    |  ,/     `\  |         |  ,6     `7  |        |  ,6     `7  | 
    |,/         `\|         |,/         `\|        |,/         `\| 
    1-------------2         1------9------2        1------9------2 

Pyramid:                     Pyramid13:                   Pyramid14:

               4                            4                            4
             ,/|\                         ,/|\                         ,/|\
           ,/ .'|\                      ,/ .'|\                      ,/ .'|\
         ,/   | | \                   ,/   | | \                   ,/   | | \
       ,/    .' | `.                ,/    .' | `.                ,/    .' | `.
     ,/      |  '.  \             ,7      |  12  \             ,7      |  12  \
   ,/       .' w |   \          ,/       .'   |   \          ,/       .'   |   \
 ,/         |  ^ |    \       ,/         9    |    11      ,/         9    |    11
0----------.'--|-3    `.     0--------6-.'----3    `.     0--------6-.'----3    `.
 `\        |   |  `\    \      `\        |      `\    \     `\        |      `\    \
   `\     .'   +----`\ - \ -> v  `5     .'        10   \      `5     .' 13     10   \
     `\   |    `\     `\  \        `\   |           `\  \       `\   |           `\  \ 
       `\.'      `\     `\`          `\.'             `\`         `\.'             `\` 
          1----------------2            1--------8-------2           1--------8-------2
                    `\
                       u
*/

element_type_t element_types[16] = {

  // 0 
 { "",                ELEMENT_TYPE_UNDEFINED,   0, 0, 0, -1, NULL, NULL, NULL, },   
  // 1
 { "line",            ELEMENT_TYPE_LINE,        1, 2, 2, 1,
         mesh_two_node_line_h,
         mesh_two_node_line_dhdr,
         mesh_point_in_line,
         mesh_line_vol,},
  // 2
 { "triangle",        ELEMENT_TYPE_TRIANGLE,    2, 3, 3, 2,
         mesh_three_node_triang_h,
         mesh_three_node_triang_dhdr,
         mesh_point_in_triangle,
         mesh_triang_vol,},
  // 3
 { "quadrangle",      ELEMENT_TYPE_QUADRANGLE,  2, 4, 4, 2,
         mesh_four_node_quad_h,
         mesh_four_node_quad_dhdr,
         mesh_point_in_quadrangle,
         mesh_quad_vol,},
  // 4
 { "tetrahedron",     ELEMENT_TYPE_TETRAHEDRON, 3, 4, 4, 3,
         mesh_four_node_tetrahedron_h,
         mesh_four_node_tetrahedron_dhdr,
         mesh_point_in_tetrahedron,
         mesh_tetrahedron_vol,},
  // 5
 { "hexahedron",      ELEMENT_TYPE_HEXAHEDRON,  3, 8, 6, 4,
         mesh_eight_node_hexahedron_h,
         mesh_eight_node_hexahedron_dhdr,
         mesh_point_in_hexahedron,
         mesh_hexahedron_vol,},
  // 6 
 { "prism",           ELEMENT_TYPE_PRISM,     3, 6, 5, 4,   
         mesh_six_node_prism_h,
         mesh_six_node_prism_dhdr,
         mesh_point_in_prism,
         mesh_prism_vol,},
  // Ojo aca que en nodos por cara pusimos el maximo valor (4) ya que depende de la cara
  // 7
 { "pyramid",         ELEMENT_TYPE_PYRAMID,   0, 0, 0, -1, NULL, NULL, NULL, },   
  // 8
 { "line3",           ELEMENT_TYPE_UNDEFINED, 0, 0, 0, -1, NULL, NULL, NULL, },   
  // 9
 { "triangle6",       ELEMENT_TYPE_UNDEFINED, 0, 0, 0, -1, NULL, NULL, NULL, },   
  // 10
 { "quadrangle9",     ELEMENT_TYPE_UNDEFINED, 0, 0, 0, -1, NULL, NULL, NULL, },   
  // 11
 { "tetrahedron10",   ELEMENT_TYPE_UNDEFINED, 0, 0, 0, -1, NULL, NULL, NULL, },   
  // 12
 { "hexahedron27",    ELEMENT_TYPE_UNDEFINED, 0, 0, 0, -1, NULL, NULL, NULL, },   
  // 13
 { "prism18",         ELEMENT_TYPE_UNDEFINED, 0, 0, 0, -1, NULL, NULL, NULL, },   
  // 14
 { "pyramid14",       ELEMENT_TYPE_UNDEFINED, 0, 0, 0, -1, NULL, NULL, NULL, },   
  // 15
 { "1-node point",    ELEMENT_TYPE_POINT,     0, 1, 0, -1,
         mesh_one_node_point_h,
         mesh_one_node_point_dhdr,
         NULL,
         mesh_point_vol,}
};


// esta no rellena los nodos!
int mesh_create_element(element_t *element, int id, int type, physical_entity_t *physical_entity) {
 
  element->id = id;
  element->type = &element_types[type];
  element->node  = calloc(element->type->nodes, sizeof(node_t *));
  element->physical_entity = physical_entity;
  
  return WASORA_RUNTIME_OK;
}

int mesh_add_element_to_list(element_list_item_t **list, element_t *element) {
  
  element_list_item_t *item = calloc(1, sizeof(element_list_item_t));
  item->element = element;
  LL_APPEND(*list, item);
  
  return WASORA_RUNTIME_OK;
  
}