/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora's mesh-related gauss integration routines
 *
 *  Copyright (C) 2014--2015 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */

#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

#include <wasora.h>
#include "mesh.h"

// inicializa puntos y pesos para cuadratura de gauss
int mesh_init_gauss_points(mesh_t *mesh, int n_gauss_points) {

  int i;
  
  if (n_gauss_points == 0) {
    n_gauss_points = 2;
  } else if (n_gauss_points < 1 || n_gauss_points > 3) {
    wasora_push_error_message("invalid number of Gauss points %d, valid values are 1, 2 or 3", mesh->n_gauss_points);
    return WASORA_RUNTIME_ERROR;
  }
  if (mesh->dimensions == 0) {
    wasora_push_error_message("mesh '%s' has no dimensions", mesh->name);
    return WASORA_RUNTIME_ERROR;
  }
 
  // karl frederich gauss
  mesh->n_gauss_points = n_gauss_points;
  mesh->n_gauss_volume = gsl_pow_int(mesh->n_gauss_points, mesh->dimensions);
  mesh->n_gauss_surface = gsl_pow_int(mesh->n_gauss_points, mesh->dimensions-1);

  // TODO: meter los allocs en init_gauss_weights
  mesh->r_gauss_volume     = calloc(mesh->dimensions, sizeof(double *));
  mesh->alpha_gauss_volume = calloc(mesh->dimensions, sizeof(double *));
  for (i = 0; i < mesh->dimensions; i++) {
    mesh->r_gauss_volume[i] = malloc(gsl_pow_int(n_gauss_points, mesh->dimensions)*sizeof(double));
    mesh->alpha_gauss_volume[i] = malloc(gsl_pow_int(n_gauss_points, mesh->dimensions)*sizeof(double));
  }

  mesh->r_gauss_surface     = calloc(mesh->dimensions-1, sizeof(double *));
  mesh->alpha_gauss_surface = calloc(mesh->dimensions-1, sizeof(double *));
  for (i = 0; i < mesh->dimensions-1; i++) {
    mesh->r_gauss_surface[i] = malloc(gsl_pow_int(n_gauss_points, mesh->dimensions-1)*sizeof(double));
    mesh->alpha_gauss_surface[i] = malloc(gsl_pow_int(n_gauss_points, mesh->dimensions-1)*sizeof(double));
  }

  wasora_call(mesh_init_gauss_weights(n_gauss_points, mesh->dimensions,   mesh->alpha_gauss_volume,  mesh->r_gauss_volume));
  wasora_call(mesh_init_gauss_weights(n_gauss_points, mesh->dimensions-1, mesh->alpha_gauss_surface, mesh->r_gauss_surface));

 
  return WASORA_RUNTIME_OK;
}


int mesh_init_gauss_weights(int n_gauss_points, int dimensions, double **alpha, double **r) {

  int i, j, k;

  double r_gauss[3][3]     = {
   { 0.0,               0,                 0},
   {-0.577350269189626, 0.577350269189626, 0},
   {-0.774596669241483, 0.0,               0.774596669241483},  
  };

  double alpha_gauss[3][3] = {
   { 2.0,     0,       0},
   { 1.0,     1.0,     0},
   { 5.0/9.0, 8.0/9.0, 5.0/9.0}
  };
  
  switch (dimensions) {
    case 1:
      for (i = 0; i < n_gauss_points; i++) {
        r[0][i] = r_gauss[n_gauss_points-1][i];
        alpha[0][i] = alpha_gauss[n_gauss_points-1][i];
      }
    break;
    case 2:
      for (i = 0; i < n_gauss_points; i++) {
        for (j = 0; j < n_gauss_points; j++) {
          r[0][j*n_gauss_points + i] = r_gauss[n_gauss_points-1][i];
          r[1][j*n_gauss_points + i] = r_gauss[n_gauss_points-1][j];

          alpha[0][j*n_gauss_points + i] = alpha_gauss[n_gauss_points-1][i];
          alpha[1][j*n_gauss_points + i] = alpha_gauss[n_gauss_points-1][j];
        }
      }
    break;
    case 3:
      for (i = 0; i < n_gauss_points; i++) {
        for (j = 0; j < n_gauss_points; j++) {
          for (k = 0; k < n_gauss_points; k++) {
            r[0][k*n_gauss_points*n_gauss_points + j*n_gauss_points + i] = r_gauss[n_gauss_points-1][i];
            r[1][k*n_gauss_points*n_gauss_points + j*n_gauss_points + i] = r_gauss[n_gauss_points-1][j];
            r[2][k*n_gauss_points*n_gauss_points + j*n_gauss_points + i] = r_gauss[n_gauss_points-1][k];

            alpha[0][k*n_gauss_points*n_gauss_points + j*n_gauss_points + i] = alpha_gauss[n_gauss_points-1][i];
            alpha[1][k*n_gauss_points*n_gauss_points + j*n_gauss_points + i] = alpha_gauss[n_gauss_points-1][j];
            alpha[2][k*n_gauss_points*n_gauss_points + j*n_gauss_points + i] = alpha_gauss[n_gauss_points-1][k];
          }
        }
      }
    break;
  }
  
  
  return WASORA_RUNTIME_OK;
}


double mesh_integral_over_element(function_t *function, element_t *element, expr_t *weight) {

  double w;
  double xi;
  double integral = 0;

  int V, v;
  int j;
  int m;
  
  mesh_t *mesh = (mesh_t *)function->mesh;
  
  V = gsl_pow_int(mesh->n_gauss_points, mesh->dimensions);
  for (v = 0; v < V; v++) {
    w = mesh_integration_weight(mesh, element, v);

    xi = 0;
    for (j = 0; j < element->type->nodes; j++) {
      xi += gsl_vector_get(mesh->fem.h, j) * function->data_value[element->node[j]->id - 1];
    }

    if (weight == NULL) {
      integral += w * xi;
    } else {
      for (m = 0; m < element->type->dim; m++) {
        gsl_vector_set(mesh->fem.r, m, mesh->r_gauss_volume[m][v]);
      }
    	mesh_compute_x(element, mesh->fem.r, wasora_value_ptr(wasora_mesh.vec_x));
      integral += wasora_evaluate_expression(weight) * w * xi;
    }
  }

  return integral;

}

/*
double mesh_integral_domain(struct function_t *f, double (*w)(double, double, double, void *), void *params) {

  int i;
  double integral;

  integral = 0;
  for (i = 0; i < mesh->n_elements; i++) {
  	if (mesh->element->type->dim == mesh->dimensions) {
      integral += mesh_integral_over_element(f, &mesh->element[i]);
  	}
  }

  return integral;

}
*/

void mesh_alloc_fem_objects(mesh_t *mesh) {
  int M, J, G, N;

  M = mesh->dimensions;
  J = mesh->max_nodes_per_element;
  G = mesh->degrees_of_freedom;
  N = J*G;

  mesh->fem.r = gsl_vector_alloc(M);
  mesh->fem.x = wasora_value_ptr(wasora_mesh.vec_x);
  mesh->fem.h = gsl_vector_alloc(J);
  mesh->fem.dhdr = gsl_matrix_alloc(J, M);
  mesh->fem.dhdx = gsl_matrix_alloc(J, M);
  mesh->fem.drdx = gsl_matrix_alloc(M, M);
  mesh->fem.dxdr = gsl_matrix_alloc(M, M);
  if (N != 0) {
    mesh->fem.l = calloc(N, sizeof(int));
  }

  return;
}

double mesh_integration_weight(mesh_t *mesh, element_t *element, int v) {
  double alpha;

  int m;

  if (mesh->fem.r == NULL) {
    mesh_alloc_fem_objects(mesh);
  }
  if (mesh->r_gauss_volume == NULL) {
    mesh_init_gauss_points(mesh, mesh->n_gauss_points);
  }

  // calculamos las coordenadas del punto de gauss y el factor de peso
  alpha = 1.0;
  if (element->type->dim == mesh->dimensions) {
    for (m = 0; m < element->type->dim; m++) {
      gsl_vector_set(mesh->fem.r, m, mesh->r_gauss_volume[m][v]);
      alpha *= mesh->alpha_gauss_volume[m][v];
    }
  } else {
    for (m = 0; m < element->type->dim; m++) {
      gsl_vector_set(mesh->fem.r, m, mesh->r_gauss_surface[m][v]);
      alpha *= mesh->alpha_gauss_surface[m][v];
    }
  }

  // evaluamos las funciones de forma en r
  mesh_compute_h(element, mesh->fem.r, mesh->fem.h);

  // dxdr
  mesh_compute_dxdr(element, mesh->fem.r, mesh->fem.dxdr);

  // calculamos el peso de integracion por el determinante del jacobiano
  return alpha * fabs(mesh_determinant(element->type->dim, mesh->fem.dxdr));
}

